"""Tests for cli.py — CLI subcommands via click CliRunner."""

from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any
from unittest.mock import MagicMock, patch

if TYPE_CHECKING:
    from pathlib import Path

from click.testing import CliRunner

from synology_mcp.cli import main


class TestCli:
    def test_version(self) -> None:
        runner = CliRunner()
        result = runner.invoke(main, ["--version"])
        assert result.exit_code == 0
        assert "0.1.0" in result.output

    def test_version_short_flag(self) -> None:
        runner = CliRunner()
        result = runner.invoke(main, ["-v"])
        assert result.exit_code == 0
        assert "0.1.0" in result.output

    def test_help(self) -> None:
        runner = CliRunner()
        result = runner.invoke(main, ["--help"])
        assert result.exit_code == 0
        assert "synology-mcp" in result.output

    def test_help_short_flag(self) -> None:
        runner = CliRunner()
        result = runner.invoke(main, ["-h"])
        assert result.exit_code == 0
        assert "synology-mcp" in result.output

    def test_serve_help(self) -> None:
        runner = CliRunner()
        result = runner.invoke(main, ["serve", "--help"])
        assert result.exit_code == 0
        assert "--config" in result.output

    def test_setup_help(self) -> None:
        runner = CliRunner()
        result = runner.invoke(main, ["setup", "--help"])
        assert result.exit_code == 0
        assert "--config" in result.output
        assert "--list" in result.output

    def test_check_help(self) -> None:
        runner = CliRunner()
        result = runner.invoke(main, ["check", "--help"])
        assert result.exit_code == 0
        assert "--config" in result.output

    def test_serve_missing_config(self) -> None:
        runner = CliRunner()
        result = runner.invoke(main, ["serve", "--config", "/nonexistent/config.yaml"])
        assert result.exit_code != 0

    def test_serve_missing_config_error_in_red(self) -> None:
        runner = CliRunner()
        result = runner.invoke(main, ["serve", "-c", "/nonexistent/config.yaml"])
        assert result.exit_code != 0
        # Error goes to stderr
        assert "not found" in (result.output + (result.stderr if hasattr(result, "stderr") else ""))

    def test_short_config_flag_serve(self) -> None:
        runner = CliRunner()
        result = runner.invoke(main, ["serve", "-c", "/nonexistent/config.yaml"])
        assert result.exit_code != 0

    def test_short_config_flag_check(self) -> None:
        runner = CliRunner()
        result = runner.invoke(main, ["check", "-c", "/nonexistent/config.yaml"])
        assert result.exit_code != 0


class TestSetupList:
    def test_list_no_configs(self, tmp_path: Path) -> None:
        runner = CliRunner()
        with patch("synology_mcp.cli._CONFIG_DIR", tmp_path / "nonexistent"):
            result = runner.invoke(main, ["setup", "--list"])
        assert result.exit_code == 0
        assert "No configurations found" in result.output

    def test_list_with_configs(self, tmp_path: Path) -> None:
        config_file = tmp_path / "my-nas.yaml"
        config_file.write_text(
            "schema_version: 1\n"
            "instance_id: my-nas\n"
            "alias: HomeNAS\n"
            "connection:\n"
            "  host: 192.168.1.100\n"
            "modules:\n"
            "  filestation:\n"
            "    enabled: true\n"
        )
        runner = CliRunner()
        with patch("synology_mcp.cli._CONFIG_DIR", tmp_path):
            result = runner.invoke(main, ["setup", "--list"])
        assert result.exit_code == 0
        assert "my-nas.yaml" in result.output
        assert "HomeNAS" in result.output
        assert "192.168.1.100" in result.output

    def test_list_short_flag(self, tmp_path: Path) -> None:
        runner = CliRunner()
        with patch("synology_mcp.cli._CONFIG_DIR", tmp_path / "nonexistent"):
            result = runner.invoke(main, ["setup", "-l"])
        assert result.exit_code == 0
        assert "No configurations found" in result.output

    def test_list_empty_directory(self, tmp_path: Path) -> None:
        """Config dir exists but has no .yaml files."""
        runner = CliRunner()
        with patch("synology_mcp.cli._CONFIG_DIR", tmp_path):
            result = runner.invoke(main, ["setup", "--list"])
        assert result.exit_code == 0
        assert "No configurations found" in result.output

    def test_list_with_unparseable_config(self, tmp_path: Path) -> None:
        """Gracefully handle a config file that can't be parsed."""
        bad_file = tmp_path / "broken.yaml"
        bad_file.write_text("{{{{invalid yaml")
        runner = CliRunner()
        with patch("synology_mcp.cli._CONFIG_DIR", tmp_path):
            result = runner.invoke(main, ["setup", "--list"])
        assert result.exit_code == 0
        assert "broken.yaml" in result.output
        assert "could not parse" in result.output

    def test_list_multiple_configs(self, tmp_path: Path) -> None:
        """Multiple config files are listed."""
        for name, host in [("nas-a.yaml", "10.0.0.1"), ("nas-b.yaml", "10.0.0.2")]:
            (tmp_path / name).write_text(
                f"schema_version: 1\nconnection:\n  host: {host}\n"
                "modules:\n  filestation:\n    enabled: true\n"
            )
        runner = CliRunner()
        with patch("synology_mcp.cli._CONFIG_DIR", tmp_path):
            result = runner.invoke(main, ["setup", "--list"])
        assert "nas-a.yaml" in result.output
        assert "nas-b.yaml" in result.output
        assert "10.0.0.1" in result.output
        assert "10.0.0.2" in result.output


class TestSetupInteractive:
    def test_interactive_setup_creates_config(self, tmp_path: Path) -> None:
        """When no config file exists, interactive mode prompts and writes a file."""
        runner = CliRunner()
        config_dir = tmp_path / "config"

        clean_env: dict[str, str] = {
            k: v for k, v in os.environ.items() if not k.startswith("SYNOLOGY_")
        }

        connect_result: dict[str, Any] = {"success": True, "hostname": "MyNAS"}

        # Input order: host, https(n), permission(read), alias(""),
        # username, password, hostname-confirm(y)
        with (
            patch("synology_mcp.cli._CONFIG_DIR", config_dir),
            patch("synology_mcp.core.config.discover_config_path", side_effect=FileNotFoundError),
            patch("synology_mcp.cli._store_keyring", return_value=True),
            patch("synology_mcp.cli.asyncio.run", return_value=connect_result),
            patch.dict(os.environ, clean_env, clear=True),
        ):
            result = runner.invoke(
                main,
                ["setup"],
                input="192.168.1.50\nn\nread\n\nadmin\npassword\ny\n",
            )

        assert "Let's create one" in result.output
        assert result.exit_code == 0
        assert (config_dir / "192-168-1-50.yaml").exists()

    def test_interactive_setup_aborts_on_overwrite_decline(self, tmp_path: Path) -> None:
        """If config file exists and user declines overwrite, abort."""
        runner = CliRunner()
        config_dir = tmp_path / "config"
        config_dir.mkdir()
        (config_dir / "192-168-1-50.yaml").write_text("old\n")

        clean_env: dict[str, str] = {
            k: v for k, v in os.environ.items() if not k.startswith("SYNOLOGY_")
        }

        connect_result: dict[str, Any] = {"success": True}

        with (
            patch("synology_mcp.cli._CONFIG_DIR", config_dir),
            patch("synology_mcp.core.config.discover_config_path", side_effect=FileNotFoundError),
            patch("synology_mcp.cli._store_keyring", return_value=True),
            patch("synology_mcp.cli.asyncio.run", return_value=connect_result),
            patch.dict(os.environ, clean_env, clear=True),
        ):
            result = runner.invoke(
                main,
                ["setup"],
                input="192.168.1.50\nn\nread\n\nadmin\npassword\nn\n",
            )

        assert "Aborted" in result.output

    def test_interactive_setup_with_https(self, tmp_path: Path) -> None:
        """HTTPS prompts for verify_ssl."""
        runner = CliRunner()
        config_dir = tmp_path / "config"

        clean_env: dict[str, str] = {
            k: v for k, v in os.environ.items() if not k.startswith("SYNOLOGY_")
        }

        connect_result: dict[str, Any] = {"success": True}

        # Prompts: host, https(y), verify_ssl(n), permission(write), alias(""),
        # username, password
        with (
            patch("synology_mcp.cli._CONFIG_DIR", config_dir),
            patch("synology_mcp.core.config.discover_config_path", side_effect=FileNotFoundError),
            patch("synology_mcp.cli._store_keyring", return_value=True),
            patch("synology_mcp.cli.asyncio.run", return_value=connect_result),
            patch.dict(os.environ, clean_env, clear=True),
        ):
            result = runner.invoke(
                main,
                ["setup"],
                # host, https(y), permission(write), alias(MyNAS), verify_ssl(n), username, password
                input="nas.local\ny\nwrite\nMyNAS\nn\nadmin\npassword\n",
            )

        assert result.exit_code == 0, result.output
        config_path = config_dir / "nas.yaml"
        assert config_path.exists()
        import yaml

        data = yaml.safe_load(config_path.read_text())
        assert data["connection"]["https"] is True
        assert data["connection"]["verify_ssl"] is False
        assert data["alias"] == "MyNAS"

    def test_interactive_setup_keyring_failure(self, tmp_path: Path) -> None:
        """When keyring fails, show env var instructions and return."""
        runner = CliRunner()
        config_dir = tmp_path / "config"

        clean_env: dict[str, str] = {
            k: v for k, v in os.environ.items() if not k.startswith("SYNOLOGY_")
        }

        with (
            patch("synology_mcp.cli._CONFIG_DIR", config_dir),
            patch("synology_mcp.core.config.discover_config_path", side_effect=FileNotFoundError),
            patch("synology_mcp.cli._store_keyring", return_value=False),
            patch.dict(os.environ, clean_env, clear=True),
        ):
            result = runner.invoke(
                main,
                ["setup"],
                input="192.168.1.50\nn\nread\n\nadmin\npassword\n",
            )

        assert result.exit_code == 0
        # Should NOT have written a config file since keyring failed
        assert not (config_dir / "192-168-1-50.yaml").exists()

    def test_interactive_setup_login_failure(self, tmp_path: Path) -> None:
        """When login fails, don't write a config file."""
        runner = CliRunner()
        config_dir = tmp_path / "config"

        clean_env: dict[str, str] = {
            k: v for k, v in os.environ.items() if not k.startswith("SYNOLOGY_")
        }

        connect_result: dict[str, Any] = {"success": False}

        with (
            patch("synology_mcp.cli._CONFIG_DIR", config_dir),
            patch("synology_mcp.core.config.discover_config_path", side_effect=FileNotFoundError),
            patch("synology_mcp.cli._store_keyring", return_value=True),
            patch("synology_mcp.cli.asyncio.run", return_value=connect_result),
            patch.dict(os.environ, clean_env, clear=True),
        ):
            result = runner.invoke(
                main,
                ["setup"],
                input="192.168.1.50\nn\nread\n\nadmin\npassword\n",
            )

        assert result.exit_code == 0
        assert not (config_dir / "192-168-1-50.yaml").exists()


class TestSetupWithConfig:
    def test_setup_with_existing_config(self, tmp_path: Path) -> None:
        """Setup with --config uses the credential flow, not interactive."""
        config_file = tmp_path / "config.yaml"
        config_file.write_text(
            "schema_version: 1\n"
            "instance_id: test-nas\n"
            "connection:\n"
            "  host: 192.168.1.100\n"
            "modules:\n"
            "  filestation:\n"
            "    enabled: true\n"
        )
        runner = CliRunner()
        with (
            patch("synology_mcp.cli._store_keyring", return_value=True),
            patch("synology_mcp.cli.asyncio.run", return_value=None),
        ):
            result = runner.invoke(
                main,
                ["setup", "-c", str(config_file)],
                input="admin\npassword\n",
            )

        assert "Setting up credentials" in result.output
        assert "test-nas" in result.output

    def test_setup_with_config_shows_display_name(self, tmp_path: Path) -> None:
        """Setup shows alias in output when available."""
        config_file = tmp_path / "config.yaml"
        config_file.write_text(
            "schema_version: 1\n"
            "instance_id: test-nas\n"
            "alias: My NAS\n"
            "connection:\n"
            "  host: 192.168.1.100\n"
            "modules:\n"
            "  filestation:\n"
            "    enabled: true\n"
        )
        runner = CliRunner()
        with (
            patch("synology_mcp.cli._store_keyring", return_value=True),
            patch("synology_mcp.cli.asyncio.run", return_value=None),
        ):
            result = runner.invoke(
                main,
                ["setup", "-c", str(config_file)],
                input="admin\npassword\n",
            )

        assert "My NAS" in result.output


class TestStoreKeyring:
    def test_store_keyring_success(self) -> None:
        from synology_mcp.cli import _store_keyring

        mock_kr = MagicMock()
        # keyring is imported inside the function, so mock at the import target
        with patch.dict("sys.modules", {"keyring": mock_kr}):
            result = _store_keyring("synology-mcp/test", "admin", "secret")

        assert result is True
        assert mock_kr.set_password.call_count == 2

    def test_store_keyring_failure(self) -> None:
        from synology_mcp.cli import _store_keyring

        mock_kr = MagicMock()
        mock_kr.set_password.side_effect = Exception("No backend")
        with patch.dict("sys.modules", {"keyring": mock_kr}):
            result = _store_keyring("synology-mcp/test", "admin", "secret")

        assert result is False


class TestEmitClaudeDesktopSnippet:
    def test_snippet_includes_dbus_on_linux(self, tmp_path: Path) -> None:
        """Interactive setup on Linux includes DBUS in the Claude Desktop snippet."""
        runner = CliRunner()
        config_dir = tmp_path / "config"

        clean_env: dict[str, str] = {
            k: v for k, v in os.environ.items() if not k.startswith("SYNOLOGY_")
        }
        clean_env["DBUS_SESSION_BUS_ADDRESS"] = "unix:path=/run/user/1000/bus"

        connect_result: dict[str, Any] = {"success": True}

        with (
            patch("synology_mcp.cli._CONFIG_DIR", config_dir),
            patch("synology_mcp.core.config.discover_config_path", side_effect=FileNotFoundError),
            patch("synology_mcp.cli._store_keyring", return_value=True),
            patch("synology_mcp.cli.asyncio.run", return_value=connect_result),
            patch.dict(os.environ, clean_env, clear=True),
            patch("sys.platform", "linux"),
        ):
            result = runner.invoke(
                main,
                ["setup"],
                input="192.168.1.50\nn\nread\n\nadmin\npassword\n",
            )

        assert result.exit_code == 0
        assert "DBUS_SESSION_BUS_ADDRESS" in result.output

    def test_snippet_no_dbus_on_macos(self, tmp_path: Path) -> None:
        """On non-Linux, no DBUS env var in the snippet."""
        runner = CliRunner()
        config_dir = tmp_path / "config"

        clean_env: dict[str, str] = {
            k: v for k, v in os.environ.items() if not k.startswith("SYNOLOGY_")
        }

        connect_result: dict[str, Any] = {"success": True}

        with (
            patch("synology_mcp.cli._CONFIG_DIR", config_dir),
            patch("synology_mcp.core.config.discover_config_path", side_effect=FileNotFoundError),
            patch("synology_mcp.cli._store_keyring", return_value=True),
            patch("synology_mcp.cli.asyncio.run", return_value=connect_result),
            patch.dict(os.environ, clean_env, clear=True),
            patch("sys.platform", "darwin"),
        ):
            result = runner.invoke(
                main,
                ["setup"],
                input="192.168.1.50\nn\nread\n\nadmin\npassword\n",
            )

        assert result.exit_code == 0
        assert "DBUS_SESSION_BUS_ADDRESS" not in result.output


class TestCheckCommand:
    def test_check_with_valid_config(self, tmp_path: Path) -> None:
        """Check command loads config and attempts login."""
        config_file = tmp_path / "config.yaml"
        config_file.write_text(
            "schema_version: 1\n"
            "instance_id: test-nas\n"
            "connection:\n"
            "  host: 192.168.1.100\n"
            "modules:\n"
            "  filestation:\n"
            "    enabled: true\n"
        )
        runner = CliRunner()
        with patch("synology_mcp.cli.asyncio.run", return_value=None):
            result = runner.invoke(main, ["check", "-c", str(config_file)])

        assert "Checking credentials" in result.output
        assert "test-nas" in result.output

    def test_check_uses_display_name(self, tmp_path: Path) -> None:
        """Check shows alias when available."""
        config_file = tmp_path / "config.yaml"
        config_file.write_text(
            "schema_version: 1\n"
            "alias: My Server\n"
            "connection:\n"
            "  host: 192.168.1.100\n"
            "modules:\n"
            "  filestation:\n"
            "    enabled: true\n"
        )
        runner = CliRunner()
        with patch("synology_mcp.cli.asyncio.run", return_value=None):
            result = runner.invoke(main, ["check", "-c", str(config_file)])

        assert "My Server" in result.output


class TestEnvVarMode:
    def test_serve_env_var_mode(self) -> None:
        """When SYNOLOGY_HOST is set and no config file, synthesize config."""
        from synology_mcp.core.config import _synthesize_env_config

        env = {"SYNOLOGY_HOST": "10.0.0.5"}
        with patch.dict(os.environ, env, clear=False):
            config = _synthesize_env_config()

        assert config is not None
        assert config.connection is not None
        assert config.connection.host == "10.0.0.5"
        assert config.instance_id == "10-0-0-5"
        assert config.modules["filestation"].permission == "read"

    def test_no_env_var_returns_none(self) -> None:
        from synology_mcp.core.config import _synthesize_env_config

        clean_env: dict[str, str] = {
            k: v for k, v in os.environ.items() if not k.startswith("SYNOLOGY_")
        }
        with patch.dict(os.environ, clean_env, clear=True):
            config = _synthesize_env_config()

        assert config is None

    def test_load_config_falls_back_to_env(self, tmp_path: Path) -> None:
        """load_config falls back to env-var mode when no config file exists."""
        from synology_mcp.core.config import load_config

        env: dict[str, str] = {
            "SYNOLOGY_HOST": "10.0.0.99",
        }
        clean_env = {k: v for k, v in os.environ.items() if not k.startswith("SYNOLOGY_")}
        clean_env.update(env)

        with patch.dict(os.environ, clean_env, clear=True):
            config = load_config(None)

        assert config is not None
        assert config.connection is not None
        assert config.connection.host == "10.0.0.99"

    def test_load_config_explicit_path_no_fallback(self) -> None:
        """Explicit --config path should not fall back to env-var mode."""
        import pytest

        from synology_mcp.core.config import load_config

        with pytest.raises(FileNotFoundError):
            load_config("/nonexistent/config.yaml")

    def test_env_var_mode_with_port_override(self) -> None:
        """Env vars can override port and https in synthesized config."""
        from synology_mcp.core.config import _synthesize_env_config

        env: dict[str, str] = {
            "SYNOLOGY_HOST": "nas.local",
            "SYNOLOGY_PORT": "5001",
            "SYNOLOGY_HTTPS": "true",
        }
        clean_env = {k: v for k, v in os.environ.items() if not k.startswith("SYNOLOGY_")}
        clean_env.update(env)

        with patch.dict(os.environ, clean_env, clear=True):
            config = _synthesize_env_config()

        assert config is not None
        assert config.connection is not None
        assert config.connection.port == 5001
        assert config.connection.https is True


class TestAliasField:
    def test_alias_in_config(self) -> None:
        raw: dict[str, Any] = {
            "schema_version": 1,
            "alias": "HomeNAS",
            "connection": {"host": "192.168.1.100"},
            "modules": {"filestation": {"enabled": True}},
        }
        from synology_mcp.core.config import AppConfig

        config = AppConfig(**raw)
        assert config.alias == "HomeNAS"
        assert config.display_name == "HomeNAS"

    def test_display_name_falls_back_to_instance_id(self) -> None:
        raw: dict[str, Any] = {
            "schema_version": 1,
            "connection": {"host": "192.168.1.100"},
            "modules": {"filestation": {"enabled": True}},
        }
        from synology_mcp.core.config import AppConfig

        config = AppConfig(**raw)
        assert config.alias is None
        assert config.display_name == "192-168-1-100"

    def test_display_name_with_alias_and_instance_id(self) -> None:
        raw: dict[str, Any] = {
            "schema_version": 1,
            "instance_id": "my-nas",
            "alias": "Office NAS",
            "connection": {"host": "10.0.0.1"},
            "modules": {"filestation": {"enabled": True}},
        }
        from synology_mcp.core.config import AppConfig

        config = AppConfig(**raw)
        assert config.display_name == "Office NAS"


class TestFetchDsmInfo:
    async def test_fetch_dsm_info_not_in_cache(self) -> None:
        """When SYNO.DSM.Info is not in the API cache, return empty dict."""
        from synology_mcp.core.client import DsmClient

        async with DsmClient(base_url="http://nas:5000") as client:
            result = await client.fetch_dsm_info()
        assert result == {}

    async def test_fetch_dsm_info_in_cache(self) -> None:
        """When SYNO.DSM.Info is available, call getinfo and return data."""
        import respx

        from synology_mcp.core.client import DsmClient
        from synology_mcp.core.state import ApiInfoEntry

        with respx.mock:
            respx.get("http://nas:5000/webapi/entry.cgi").respond(
                json={
                    "success": True,
                    "data": {
                        "model": "DS1618+",
                        "hostname": "MyNAS",
                        "version_string": "DSM 7.1.1-42962 Update 6",
                    },
                }
            )
            async with DsmClient(base_url="http://nas:5000") as client:
                client._api_cache = {
                    "SYNO.DSM.Info": ApiInfoEntry(path="entry.cgi", min_version=1, max_version=2),
                }
                result = await client.fetch_dsm_info()

        assert result["hostname"] == "MyNAS"
        assert result["model"] == "DS1618+"



"""CLI: serve, setup, check subcommands (click-based)."""

from __future__ import annotations

import asyncio
import json as json_mod
import logging
import os
import subprocess
import sys
from datetime import datetime
from pathlib import Path
from typing import Any
from urllib.request import Request, urlopen

import click
import yaml

from synology_mcp import __version__

_LOG_FORMAT = "%(asctime)s [%(levelname)s] %(name)s: %(message)s"

_CONFIG_DIR = Path.home() / ".config" / "synology-mcp"

_PYPI_PACKAGE = "synology-mcp"
_VERSION_CHECK_INTERVAL = 86400  # 24 hours


# --- Version checking and upgrades ---


def _get_current_version() -> str:
    """Return the running version."""
    try:
        from importlib.metadata import version as _pkg_version

        return _pkg_version(_PYPI_PACKAGE)
    except Exception:  # noqa: BLE001
        return __version__


def _get_latest_pypi_version() -> str | None:
    """Fetch the latest version from PyPI. Returns None on any failure."""
    try:
        req = Request(
            f"https://pypi.org/pypi/{_PYPI_PACKAGE}/json",
            headers={"User-Agent": f"{_PYPI_PACKAGE}/{_get_current_version()}"},
        )
        with urlopen(req, timeout=5) as resp:  # noqa: S310
            data = json_mod.loads(resp.read())
        result: str = data["info"]["version"]
        return result
    except Exception:  # noqa: BLE001
        return None


def _version_tuple(v: str) -> tuple[int, ...]:
    try:
        return tuple(int(x) for x in v.split("."))
    except Exception:  # noqa: BLE001
        return (0,)


def _detect_installer() -> str | None:
    """Detect whether the tool was installed via uv tool or pipx."""
    import shutil

    bin_path = shutil.which(_PYPI_PACKAGE) or ""
    if "uv" in bin_path and "tools" in bin_path:
        return "uv"
    if "pipx" in bin_path:
        return "pipx"
    return None


def _load_global_state() -> dict[str, Any]:
    """Load global (non-instance) state for version tracking."""
    path = Path.home() / ".local" / "state" / "synology-mcp" / "global.yaml"
    if not path.exists():
        return {}
    try:
        raw: dict[str, Any] = yaml.safe_load(path.read_text(encoding="utf-8")) or {}
        return raw
    except Exception:  # noqa: BLE001
        return {}


def _save_global_state(state: dict[str, Any]) -> None:
    """Save global state."""
    path = Path.home() / ".local" / "state" / "synology-mcp" / "global.yaml"
    path.parent.mkdir(parents=True, exist_ok=True)
    raw_text = yaml.dump(state, default_flow_style=False, sort_keys=False)
    path.write_text("# Auto-generated by synology-mcp.\n" + raw_text, encoding="utf-8")


def _check_for_update(state: dict[str, Any], *, force: bool = False) -> str | None:
    """Check PyPI for a newer version. Returns latest version string if newer, else None."""
    now = datetime.now().timestamp()
    if not force:
        last = state.get("last_version_check")
        if last:
            try:
                elapsed = now - datetime.fromisoformat(last).timestamp()
                if elapsed < _VERSION_CHECK_INTERVAL:
                    latest: str | None = state.get("latest_known_version")
                    current = _get_current_version()
                    if latest and _version_tuple(latest) > _version_tuple(current):
                        return latest
                    return None
            except Exception:  # noqa: BLE001
                pass

    latest = _get_latest_pypi_version()
    state["last_version_check"] = datetime.now().isoformat()
    if not latest:
        return None

    state["latest_known_version"] = latest
    current = _get_current_version()

    if _version_tuple(latest) > _version_tuple(current):
        return latest
    return None


def _do_auto_upgrade(state: dict[str, Any]) -> bool:
    """Upgrade to the latest version. Returns True on success."""
    installer = _detect_installer()
    current = _get_current_version()

    if installer == "uv":
        cmd = ["uv", "tool", "install", f"{_PYPI_PACKAGE}@latest"]
    elif installer == "pipx":
        cmd = ["pipx", "upgrade", _PYPI_PACKAGE]
    else:
        click.echo(
            "Cannot detect package manager (uv or pipx). "
            f"Upgrade manually: uv tool upgrade {_PYPI_PACKAGE}"
        )
        return False

    click.echo(f"Upgrading {_PYPI_PACKAGE} via {installer}...")
    result = subprocess.run(cmd, capture_output=True, text=True)  # noqa: S603
    if result.returncode == 0:
        state["previous_version"] = current
        _save_global_state(state)
        click.echo(click.style("Upgrade complete — takes effect on next run.", fg="green"))
        return True
    click.echo(click.style(f"Upgrade failed: {result.stderr.strip()}", fg="red"), err=True)
    return False


def _do_revert(target_version: str | None = None) -> None:
    """Revert to a specific version or the previously recorded one."""
    state = _load_global_state()
    current = _get_current_version()

    prev: str | None
    if target_version:
        prev = target_version
    else:
        prev = state.get("previous_version")
        if not prev:
            click.echo("No previous version recorded — nothing to revert to.")
            click.echo(f"Tip: specify a version explicitly: {_PYPI_PACKAGE} --revert 0.1.0")
            return

    if prev == current:
        click.echo(f"Already running {current} — nothing to do.")
        return

    installer = _detect_installer()

    if installer == "uv":
        cmd = ["uv", "tool", "install", "--force", f"{_PYPI_PACKAGE}=={prev}"]
    elif installer == "pipx":
        cmd = ["pipx", "install", "--force", f"{_PYPI_PACKAGE}=={prev}"]
    else:
        click.echo(
            f"Cannot detect package manager. Install manually:\n"
            f"  uv tool install --force {_PYPI_PACKAGE}=={prev}\n"
            f"  pipx install --force {_PYPI_PACKAGE}=={prev}"
        )
        return

    click.echo(f"Reverting from {current} to {prev}...")
    result = subprocess.run(cmd, capture_output=True, text=True)  # noqa: S603
    if result.returncode == 0:
        state["previous_version"] = None
        state["auto_upgrade"] = False
        _save_global_state(state)
        click.echo(click.style(f"Reverted to {prev}. Takes effect on next run.", fg="green"))
        click.echo("Auto-upgrade has been disabled to prevent immediately re-upgrading.")
        click.echo(f"Re-enable with: {_PYPI_PACKAGE} --auto-upgrade enable")
    else:
        click.echo(click.style(f"Revert failed: {result.stderr.strip()}", fg="red"), err=True)


def _init_early_logging(*, verbose: bool = False) -> None:
    """Configure logging early so config discovery/loading is visible.

    Checks --verbose flag first, then SYNOLOGY_LOG_LEVEL env var.
    Falls back to INFO. Will be reconfigured after config is loaded
    if the config specifies a different level.
    """
    if verbose:
        level = logging.DEBUG
    else:
        env_level = os.environ.get("SYNOLOGY_LOG_LEVEL", "info").upper()
        level = getattr(logging, env_level, logging.INFO)
    logging.basicConfig(level=level, format=_LOG_FORMAT, stream=sys.stderr)


def _configure_logging(level: str, log_file: str | None = None) -> None:
    """Reconfigure logging from loaded config values."""
    root = logging.getLogger()
    root.setLevel(getattr(logging, level.upper(), logging.INFO))

    if log_file:
        file_handler = logging.FileHandler(os.path.expanduser(log_file), encoding="utf-8")
        file_handler.setFormatter(logging.Formatter(_LOG_FORMAT))
        root.addHandler(file_handler)


@click.group(context_settings={"help_option_names": ["-h", "--help"]}, invoke_without_command=True)
@click.version_option(__version__, "-v", "--version", prog_name="synology-mcp")
@click.option("--check-update", is_flag=True, help="Check PyPI for a newer version")
@click.option(
    "--auto-upgrade",
    type=click.Choice(["enable", "disable"]),
    help="Enable or disable automatic upgrades",
)
@click.option(
    "--revert",
    nargs=1,
    required=False,
    default=None,
    metavar="[VERSION]",
    is_flag=False,
    flag_value="__PREVIOUS__",
    help="Revert to previous or specified version",
)
@click.pass_context
def main(
    ctx: click.Context,
    check_update: bool,
    auto_upgrade: str | None,
    revert: str | None,
) -> None:
    """synology-mcp — MCP server for Synology NAS."""
    if check_update:
        state = _load_global_state()
        latest = _check_for_update(state, force=True)
        _save_global_state(state)
        current = _get_current_version()
        if latest:
            click.echo(f"Update available: {current} -> {latest}")
            installer = _detect_installer()
            if installer == "uv":
                click.echo(f"Upgrade with: uv tool install {_PYPI_PACKAGE}@latest")
            elif installer == "pipx":
                click.echo(f"Upgrade with: pipx upgrade {_PYPI_PACKAGE}")
            else:
                click.echo(f"Upgrade with: uv tool install {_PYPI_PACKAGE}@latest")
        else:
            click.echo(f"You are running the latest version ({current}).")
        ctx.exit()

    if auto_upgrade is not None:
        state = _load_global_state()
        state["auto_upgrade"] = auto_upgrade == "enable"
        _save_global_state(state)
        status = "enabled" if state["auto_upgrade"] else "disabled"
        click.echo(f"Auto-upgrade {status}.")
        ctx.exit()

    if revert is not None:
        _do_revert(None if revert == "__PREVIOUS__" else revert)
        ctx.exit()

    # Track version changes for --revert
    state = _load_global_state()
    current = _get_current_version()
    last_known = state.get("running_version")
    if last_known and last_known != current:
        state["previous_version"] = last_known
    state["running_version"] = current
    _save_global_state(state)

    # Auto-upgrade check — only on interactive commands, not serve
    # (serve is launched by Claude Desktop; upgrading mid-launch is risky)
    if state.get("auto_upgrade") and ctx.invoked_subcommand != "serve":
        latest = _check_for_update(state)
        _save_global_state(state)
        if latest:
            _do_auto_upgrade(state)

    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())


@main.command()
@click.option("-c", "--config", type=click.Path(), help="Path to config file")
def serve(config: str | None) -> None:
    """Start the MCP server (launched by Claude Desktop)."""
    _init_early_logging()

    from synology_mcp.core.config import load_config
    from synology_mcp.server import create_server

    try:
        app_config = load_config(config)
    except (FileNotFoundError, ValueError) as e:
        click.echo(click.style(f"Error: {e}", fg="red"), err=True)
        sys.exit(1)

    _configure_logging(app_config.logging.level, app_config.logging.file)

    server = create_server(app_config)
    server.run(transport="stdio")


@main.command()
@click.option("-c", "--config", type=click.Path(), help="Path to config file")
@click.option("-l", "--list", "list_configs", is_flag=True, help="List existing configurations")
@click.option("--verbose", is_flag=True, help="Enable debug logging")
def setup(config: str | None, list_configs: bool, verbose: bool) -> None:
    """Interactive credential setup and 2FA bootstrap."""
    _init_early_logging(verbose=verbose)

    if list_configs:
        _list_configurations()
        return

    # If --config is given, use the existing config-file flow
    if config:
        _setup_with_config(config, verbose)
        return

    # Try to load an existing config via discovery
    from synology_mcp.core.config import load_config

    try:
        app_config = load_config(None)
    except FileNotFoundError:
        # No config file found — enter interactive creation mode
        _setup_interactive(verbose)
        return
    except ValueError as e:
        click.echo(click.style(f"Error: {e}", fg="red"), err=True)
        sys.exit(1)

    # Config file exists — do credential setup against it
    _setup_credential_flow(app_config, verbose)


def _list_configurations() -> None:
    """List existing configuration files in the config directory."""
    if not _CONFIG_DIR.exists():
        click.echo("No configurations found.")
        click.echo(f"Config directory: {_CONFIG_DIR}")
        return

    yaml_files = sorted(_CONFIG_DIR.glob("*.yaml"))
    if not yaml_files:
        click.echo("No configurations found.")
        click.echo(f"Config directory: {_CONFIG_DIR}")
        return

    click.echo(f"Configurations in {_CONFIG_DIR}:\n")
    for f in yaml_files:
        try:
            raw = yaml.safe_load(f.read_text(encoding="utf-8")) or {}
            host = raw.get("connection", {}).get("host", "?")
            alias = raw.get("alias", "")
            instance_id = raw.get("instance_id", f.stem)
            label = alias or instance_id
            click.echo(f"  {f.name}")
            click.echo(f"    Name: {label}")
            click.echo(f"    Host: {host}")
            click.echo()
        except Exception:  # noqa: BLE001
            click.echo(f"  {f.name}  (could not parse)")


def _setup_interactive(verbose: bool) -> None:
    """Interactive setup when no config file exists."""
    click.echo("No configuration file found. Let's create one.\n")

    # Prompt for connection details
    host = click.prompt("NAS host (IP or hostname)")
    use_https = click.confirm("Use HTTPS?", default=False)
    permission = click.prompt(
        "File Station permission",
        type=click.Choice(["read", "write"]),
        default="read",
    )

    # Derive instance_id for file naming
    from synology_mcp.core.config import _derive_instance_id

    instance_id = _derive_instance_id(host)

    # Prompt for alias
    alias_input = click.prompt("Alias (friendly name, optional)", default="", show_default=False)
    alias: str | None = alias_input.strip() or None

    # Build config dict
    conn: dict[str, Any] = {"host": host}
    if use_https:
        conn["https"] = True
        conn["verify_ssl"] = click.confirm("Verify SSL certificate?", default=False)

    config_dict: dict[str, Any] = {
        "schema_version": 1,
        "instance_id": instance_id,
        "connection": conn,
        "modules": {
            "filestation": {
                "enabled": True,
                "permission": permission,
            }
        },
    }
    if alias:
        config_dict["alias"] = alias

    # Validate before writing
    from synology_mcp.core.config import AppConfig

    try:
        app_config = AppConfig(**config_dict)
    except ValueError as e:
        click.echo(click.style(f"Config validation failed: {e}", fg="red"), err=True)
        sys.exit(1)

    # Prompt for credentials
    click.echo()
    username = click.prompt("DSM username")
    password = click.prompt("DSM password", hide_input=True)

    # Store in keyring
    service = f"synology-mcp/{instance_id}"
    keyring_ok = _store_keyring(service, username, password)
    if not keyring_ok:
        return

    # Connect, validate login, fetch hostname
    click.echo("\nConnecting to NAS...")
    result = asyncio.run(_connect_and_login(app_config, username, password, service, verbose))
    if not result["success"]:
        return

    # Suggest hostname as alias if we fetched one and user didn't set an alias
    nas_hostname = result.get("hostname")
    if nas_hostname and not alias:
        use_hostname = click.confirm(
            f'NAS reports hostname "{nas_hostname}". Use as alias?', default=True
        )
        if use_hostname:
            alias = nas_hostname
            config_dict["alias"] = alias
            # Re-validate
            app_config = AppConfig(**config_dict)

    # Check for existing file before writing
    config_path = _CONFIG_DIR / f"{instance_id}.yaml"
    if config_path.exists() and not click.confirm(
        f"\n{config_path} already exists. Overwrite?", default=False
    ):
        click.echo("Aborted.")
        return

    # Write config file
    _CONFIG_DIR.mkdir(parents=True, exist_ok=True)
    raw_yaml = yaml.dump(config_dict, default_flow_style=False, sort_keys=False)
    header = "# Generated by synology-mcp setup\n"
    config_path.write_text(header + raw_yaml, encoding="utf-8")
    click.echo(f"\nConfig written to {config_path}")

    # Emit Claude Desktop JSON snippet
    _emit_claude_desktop_snippet(app_config, config_path)


def _setup_with_config(config_path_str: str, verbose: bool) -> None:
    """Setup using an explicit config file path."""
    from synology_mcp.core.config import load_config

    try:
        app_config = load_config(config_path_str)
    except (FileNotFoundError, ValueError) as e:
        click.echo(click.style(f"Error: {e}", fg="red"), err=True)
        sys.exit(1)

    _setup_credential_flow(app_config, verbose)


def _setup_credential_flow(app_config: Any, verbose: bool) -> None:
    """Credential setup flow for an existing config."""
    from synology_mcp.core.config import AppConfig

    assert isinstance(app_config, AppConfig)

    if not verbose:
        _configure_logging(app_config.logging.level, app_config.logging.file)

    assert app_config.connection is not None
    click.echo(f"Setting up credentials for {app_config.display_name}")
    click.echo(f"  Host: {app_config.connection.host}")
    click.echo(f"  Instance: {app_config.instance_id}")
    click.echo()

    username = click.prompt("DSM username")
    password = click.prompt("DSM password", hide_input=True)

    service = f"synology-mcp/{app_config.instance_id}"
    keyring_ok = _store_keyring(service, username, password)
    if not keyring_ok:
        return

    click.echo("\nAttempting login...")
    asyncio.run(_setup_login(app_config, username, password, service))


def _store_keyring(service: str, username: str, password: str) -> bool:
    """Store credentials in the OS keyring. Returns True on success."""
    try:
        import keyring

        keyring.set_password(service, "username", username)
        keyring.set_password(service, "password", password)
        click.echo("\nCredentials stored in OS keyring.")
        return True
    except Exception:  # noqa: BLE001
        click.echo(
            "\nKeyring not available. Set environment variables instead:\n"
            f"  export SYNOLOGY_USERNAME={username}\n"
            f"  export SYNOLOGY_PASSWORD=<your-password>"
        )
        return False


async def _connect_and_login(
    config: Any, username: str, password: str, service: str, verbose: bool
) -> dict[str, Any]:
    """Connect to NAS, validate login, fetch hostname. Returns result dict."""
    from synology_mcp.core.client import DsmClient
    from synology_mcp.core.config import AppConfig

    assert isinstance(config, AppConfig)
    assert config.connection is not None

    protocol = "https" if config.connection.https else "http"
    base_url = f"{protocol}://{config.connection.host}:{config.connection.port}"

    result: dict[str, Any] = {"success": False}

    async with DsmClient(
        base_url=base_url,
        verify_ssl=config.connection.verify_ssl,
        timeout=config.connection.timeout,
    ) as client:
        await client.query_api_info()

        # Attempt login
        try:
            data = await client.request(
                "SYNO.API.Auth",
                "login",
                version=6,
                params={
                    "account": username,
                    "passwd": password,
                    "format": "sid",
                },
            )
        except Exception as e:  # noqa: BLE001
            error_code = getattr(e, "code", None)
            if error_code == 403:
                # 2FA flow
                click.echo("2FA is required. Enter the OTP code from your authenticator app.")
                otp_code = click.prompt("OTP code")
                try:
                    data = await client.request(
                        "SYNO.API.Auth",
                        "login",
                        version=6,
                        params={
                            "account": username,
                            "passwd": password,
                            "otp_code": otp_code,
                            "enable_device_token": "yes",
                            "device_name": "SynologyMCP",
                            "format": "sid",
                        },
                    )
                    device_id = data.get("did", "")
                    if device_id:
                        import keyring

                        keyring.set_password(service, "device_id", device_id)
                        click.echo(click.style("2FA bootstrap complete!", fg="green"))
                        click.echo("Device token stored in keyring.")
                    else:
                        click.echo(click.style("Login successful!", fg="green"))
                except Exception as e2:  # noqa: BLE001
                    click.echo(click.style(f"Login failed: {e2}", fg="red"), err=True)
                    return result
            else:
                click.echo(click.style(f"Login failed: {e}", fg="red"), err=True)
                return result

        sid = data.get("sid")
        if sid:
            client.sid = sid
            click.echo(click.style("Login successful!", fg="green"))

            # Fetch NAS hostname
            dsm_info = await client.fetch_dsm_info()
            nas_hostname = dsm_info.get("hostname") or ""
            if nas_hostname:
                result["hostname"] = nas_hostname

            # Store DSM version in state
            dsm_version = dsm_info.get("version_string", "")
            if dsm_version:
                result["dsm_version"] = dsm_version

            # Logout
            await client.request("SYNO.API.Auth", "logout", version=6, params={})

        result["success"] = True

    return result


async def _setup_login(config: object, username: str, password: str, service: str) -> None:
    """Attempt login during setup, handle 2FA if needed."""
    from synology_mcp.core.config import AppConfig

    assert isinstance(config, AppConfig)
    assert config.connection is not None

    from synology_mcp.core.client import DsmClient

    protocol = "https" if config.connection.https else "http"
    base_url = f"{protocol}://{config.connection.host}:{config.connection.port}"

    async with DsmClient(
        base_url=base_url,
        verify_ssl=config.connection.verify_ssl,
        timeout=config.connection.timeout,
    ) as client:
        await client.query_api_info()

        # Attempt login
        try:
            data = await client.request(
                "SYNO.API.Auth",
                "login",
                version=6,
                params={
                    "account": username,
                    "passwd": password,
                    "format": "sid",
                },
            )
            sid = data.get("sid")
            if sid:
                click.echo(click.style("Login successful!", fg="green"))
                # Logout
                client.sid = sid
                await client.request("SYNO.API.Auth", "logout", version=6, params={})
        except Exception as e:  # noqa: BLE001
            error_code = getattr(e, "code", None)
            if error_code == 403:
                click.echo("2FA is required. Enter the OTP code from your authenticator app.")
                otp_code = click.prompt("OTP code")

                try:
                    data = await client.request(
                        "SYNO.API.Auth",
                        "login",
                        version=6,
                        params={
                            "account": username,
                            "passwd": password,
                            "otp_code": otp_code,
                            "enable_device_token": "yes",
                            "device_name": "SynologyMCP",
                            "format": "sid",
                        },
                    )
                    device_id = data.get("did", "")
                    if device_id:
                        import keyring

                        keyring.set_password(service, "device_id", device_id)
                        click.echo(click.style("2FA bootstrap complete!", fg="green"))
                        click.echo("Device token stored in keyring.")
                    else:
                        click.echo(click.style("Login successful!", fg="green"))

                    sid = data.get("sid")
                    if sid:
                        client.sid = sid
                        await client.request("SYNO.API.Auth", "logout", version=6, params={})
                except Exception as e2:  # noqa: BLE001
                    click.echo(click.style(f"Login failed: {e2}", fg="red"), err=True)
            else:
                click.echo(click.style(f"Login failed: {e}", fg="red"), err=True)


def _emit_claude_desktop_snippet(config: Any, config_path: Path) -> None:
    """Print a Claude Desktop JSON snippet for the user to copy."""
    import json
    import shutil

    uv_path = shutil.which("uv") or "<path-to-uv>"

    server_entry: dict[str, Any] = {
        "command": uv_path,
        "args": [
            "--directory",
            str(Path.cwd()),
            "run",
            "synology-mcp",
            "serve",
            "--config",
            str(config_path),
        ],
    }

    # On Linux, include DBUS_SESSION_BUS_ADDRESS so the server process
    # can access the OS keyring (GNOME Keyring / KWallet via D-Bus).
    # Use the env var if set, otherwise construct the standard systemd path.
    if sys.platform == "linux":
        dbus_addr = os.environ.get("DBUS_SESSION_BUS_ADDRESS")
        if not dbus_addr:
            dbus_addr = f"unix:path=/run/user/{os.getuid()}/bus"
        server_entry["env"] = {"DBUS_SESSION_BUS_ADDRESS": dbus_addr}

    snippet = {"mcpServers": {f"synology-{config.display_name}": server_entry}}
    click.echo("\nAdd this to your Claude Desktop config (claude_desktop_config.json):\n")
    click.echo(json.dumps(snippet, indent=2))


@main.command()
@click.option("-c", "--config", type=click.Path(), help="Path to config file")
@click.option("--verbose", is_flag=True, help="Enable debug logging")
def check(config: str | None, verbose: bool) -> None:
    """Validate stored credentials can authenticate."""
    _init_early_logging(verbose=verbose)

    from synology_mcp.core.config import load_config

    try:
        app_config = load_config(config)
    except (FileNotFoundError, ValueError) as e:
        click.echo(click.style(f"Error: {e}", fg="red"), err=True)
        sys.exit(1)

    if not verbose:
        _configure_logging(app_config.logging.level, app_config.logging.file)

    click.echo(f"Checking credentials for '{app_config.display_name}'...")
    asyncio.run(_check_login(app_config))


async def _check_login(config: object) -> None:
    """Validate credentials by attempting a login."""
    from synology_mcp.core.auth import AuthManager
    from synology_mcp.core.client import DsmClient
    from synology_mcp.core.config import AppConfig

    assert isinstance(config, AppConfig)
    assert config.connection is not None

    protocol = "https" if config.connection.https else "http"
    base_url = f"{protocol}://{config.connection.host}:{config.connection.port}"

    async with DsmClient(
        base_url=base_url,
        verify_ssl=config.connection.verify_ssl,
        timeout=config.connection.timeout,
    ) as client:
        await client.query_api_info()

        auth = AuthManager(config, client)
        try:
            await auth.login()
            click.echo(click.style("Authentication successful!", fg="green"))
            await auth.logout()
        except Exception as e:  # noqa: BLE001
            click.echo(click.style(f"Authentication failed: {e}", fg="red"), err=True)
            sys.exit(1)



"""Tests for modules/__init__.py — module system, permission tiers, versioned handlers."""

from __future__ import annotations

import pytest

from synology_mcp.core.state import ApiInfoEntry
from synology_mcp.modules import (
    ApiRequirement,
    ModuleInfo,
    PermissionTier,
    ToolInfo,
    VersionedHandler,
    filter_tools_by_permission,
    validate_api_requirements,
)


class TestPermissionTier:
    def test_ordering(self) -> None:
        assert PermissionTier.READ < PermissionTier.WRITE
        assert PermissionTier.WRITE < PermissionTier.ADMIN
        assert PermissionTier.ADMIN > PermissionTier.READ

    def test_equality(self) -> None:
        assert PermissionTier.READ <= PermissionTier.READ
        assert PermissionTier.READ >= PermissionTier.READ

    def test_values(self) -> None:
        assert PermissionTier.READ.value == "read"
        assert PermissionTier.WRITE.value == "write"
        assert PermissionTier.ADMIN.value == "admin"


class TestFilterToolsByPermission:
    def _sample_tools(self) -> list[ToolInfo]:
        return [
            ToolInfo("list_shares", "List shares", PermissionTier.READ),
            ToolInfo("list_files", "List files", PermissionTier.READ),
            ToolInfo("move_files", "Move files", PermissionTier.WRITE),
            ToolInfo("delete_files", "Delete files", PermissionTier.WRITE),
            ToolInfo("admin_op", "Admin operation", PermissionTier.ADMIN),
        ]

    def test_read_tier(self) -> None:
        tools = self._sample_tools()
        allowed = filter_tools_by_permission(tools, PermissionTier.READ)
        assert allowed == {"list_shares", "list_files"}

    def test_write_tier(self) -> None:
        tools = self._sample_tools()
        allowed = filter_tools_by_permission(tools, PermissionTier.WRITE)
        assert allowed == {"list_shares", "list_files", "move_files", "delete_files"}

    def test_admin_tier(self) -> None:
        tools = self._sample_tools()
        allowed = filter_tools_by_permission(tools, PermissionTier.ADMIN)
        assert len(allowed) == 5


class TestValidateApiRequirements:
    def _sample_cache(self) -> dict[str, ApiInfoEntry]:
        return {
            "SYNO.FileStation.List": ApiInfoEntry(path="entry.cgi", min_version=1, max_version=2),
            "SYNO.FileStation.Search": ApiInfoEntry(path="entry.cgi", min_version=1, max_version=2),
        }

    def test_all_requirements_met(self) -> None:
        reqs = [
            ApiRequirement("SYNO.FileStation.List", min_version=1),
            ApiRequirement("SYNO.FileStation.Search", min_version=1),
        ]
        issues = validate_api_requirements(reqs, self._sample_cache())
        assert issues == []

    def test_missing_required_api(self) -> None:
        reqs = [ApiRequirement("SYNO.NonExistent", min_version=1)]
        issues = validate_api_requirements(reqs, self._sample_cache())
        assert len(issues) == 1
        assert "not available" in issues[0]

    def test_missing_optional_api(self) -> None:
        reqs = [ApiRequirement("SYNO.NonExistent", min_version=1, optional=True)]
        issues = validate_api_requirements(reqs, self._sample_cache())
        assert len(issues) == 1
        assert "Optional" in issues[0]

    def test_version_too_high(self) -> None:
        reqs = [ApiRequirement("SYNO.FileStation.List", min_version=5)]
        issues = validate_api_requirements(reqs, self._sample_cache())
        assert len(issues) == 1
        assert "v5" in issues[0]


class TestVersionedHandler:
    async def test_single_version(self) -> None:
        handler = VersionedHandler()

        @handler.version(1, 2)
        async def handle_v1(**kwargs: object) -> str:
            return "v1-2"

        result = await handler(negotiated_version=1)
        assert result == "v1-2"

    async def test_dispatch_to_best_match(self) -> None:
        handler = VersionedHandler()

        @handler.version(1, 1)
        async def handle_v1(**kwargs: object) -> str:
            return "v1"

        @handler.version(2, 3)
        async def handle_v2_v3(**kwargs: object) -> str:
            return "v2-3"

        assert await handler(negotiated_version=1) == "v1"
        assert await handler(negotiated_version=2) == "v2-3"
        assert await handler(negotiated_version=3) == "v2-3"

    async def test_no_matching_handler(self) -> None:
        handler = VersionedHandler()

        @handler.version(1, 2)
        async def handle_v1(**kwargs: object) -> str:
            return "v1"

        with pytest.raises(ValueError, match="No handler"):
            await handler(negotiated_version=5)

    async def test_passes_kwargs(self) -> None:
        handler = VersionedHandler()

        @handler.version(1)
        async def handle_v1(path: str = "") -> str:
            return f"got:{path}"

        result = await handler(negotiated_version=1, path="/video")
        assert result == "got:/video"


class TestModuleInfo:
    def test_module_info_creation(self) -> None:
        info = ModuleInfo(
            name="filestation",
            description="File Station module",
            required_apis=[ApiRequirement("SYNO.FileStation.List")],
            tools=[ToolInfo("list_shares", "List shares")],
        )
        assert info.name == "filestation"
        assert len(info.tools) == 1
        assert info.settings_schema is None



"""State file read/write.

Server-managed runtime state stored at:
  ~/.local/state/synology-mcp/{instance_id}/state.yaml
"""

from __future__ import annotations

from pathlib import Path
from typing import Any

import yaml
from pydantic import BaseModel, Field


class ApiInfoEntry(BaseModel):
    """Cached API endpoint info from SYNO.API.Info."""

    path: str
    min_version: int
    max_version: int
    request_format: str | None = None


class ServerState(BaseModel):
    """Persistent server state."""

    api_info_cache: dict[str, ApiInfoEntry] = Field(default_factory=dict)
    negotiated_versions: dict[str, int] = Field(default_factory=dict)
    recycle_bin_status: dict[str, bool] = Field(default_factory=dict)
    last_connected: str | None = None
    dsm_version: str | None = None
    hostname: str | None = None


def _state_path(instance_id: str) -> Path:
    """Get the state file path for an instance."""
    return Path.home() / ".local" / "state" / "synology-mcp" / instance_id / "state.yaml"


def load_state(instance_id: str) -> ServerState:
    """Load state from disk. Returns default state if file doesn't exist."""
    path = _state_path(instance_id)
    if not path.exists():
        return ServerState()

    raw_text = path.read_text(encoding="utf-8")
    raw: dict[str, Any] = yaml.safe_load(raw_text) or {}
    return ServerState(**raw)


def save_state(instance_id: str, state: ServerState) -> None:
    """Save state to disk. Creates directories if needed."""
    path = _state_path(instance_id)
    path.parent.mkdir(parents=True, exist_ok=True)

    data = state.model_dump(mode="json", exclude_none=True)
    raw_text = yaml.dump(data, default_flow_style=False, sort_keys=False)

    # Add header comment
    header = "# Auto-generated by synology-mcp. Do not edit.\n"
    path.write_text(header + raw_text, encoding="utf-8")



"""Tests for core/config.py — config loading, validation, env var merging."""

from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any
from unittest.mock import patch

import pytest

from synology_mcp.core.config import (
    AppConfig,
    ConnectionConfig,
    LoggingConfig,
    ModuleConfig,
    _derive_instance_id,
    _merge_env_overrides,
    discover_config_path,
    load_config,
)

if TYPE_CHECKING:
    from pathlib import Path


def _minimal_raw() -> dict[str, Any]:
    return {
        "schema_version": 1,
        "connection": {"host": "192.168.1.100"},
        "modules": {"filestation": {"enabled": True}},
    }


class TestAppConfig:
    def test_minimal_config(self) -> None:
        config = AppConfig(**_minimal_raw())
        assert config.connection is not None
        assert config.connection.host == "192.168.1.100"
        assert config.connection.port == 5000  # default for http
        assert config.instance_id == "192-168-1-100"

    def test_full_config(self) -> None:
        raw: dict[str, Any] = {
            "schema_version": 1,
            "instance_id": "nas-primary",
            "connection": {
                "host": "nas.local",
                "port": 5001,
                "https": True,
                "verify_ssl": False,
                "timeout": 60,
            },
            "auth": {"username": "admin"},
            "modules": {
                "filestation": {
                    "enabled": True,
                    "permission": "write",
                    "settings": {"async_timeout": 180},
                },
            },
            "logging": {"level": "debug", "file": "/tmp/test.log"},
        }
        config = AppConfig(**raw)
        assert config.instance_id == "nas-primary"
        assert config.connection is not None
        assert config.connection.https is True
        assert config.modules["filestation"].permission == "write"

    def test_https_default_port(self) -> None:
        raw = _minimal_raw()
        raw["connection"]["https"] = True
        config = AppConfig(**raw)
        assert config.connection is not None
        assert config.connection.port == 5001

    def test_instance_id_from_ip(self) -> None:
        config = AppConfig(**_minimal_raw())
        assert config.instance_id == "192-168-1-100"

    def test_instance_id_from_hostname(self) -> None:
        raw = _minimal_raw()
        raw["connection"]["host"] = "nas.local"
        config = AppConfig(**raw)
        assert config.instance_id == "nas"

    def test_wrong_schema_version(self) -> None:
        raw = _minimal_raw()
        raw["schema_version"] = 99
        with pytest.raises(ValueError, match="schema_version"):
            AppConfig(**raw)

    def test_missing_modules(self) -> None:
        raw = _minimal_raw()
        raw["modules"] = {}
        with pytest.raises(ValueError, match="module"):
            AppConfig(**raw)

    def test_missing_connection_host(self) -> None:
        raw: dict[str, Any] = {
            "schema_version": 1,
            "modules": {"filestation": {"enabled": True}},
        }
        with pytest.raises(ValueError, match="host"):
            AppConfig(**raw)

    def test_unknown_top_level_key_rejected(self) -> None:
        raw = _minimal_raw()
        raw["typo_key"] = "oops"
        with pytest.raises(ValueError):
            AppConfig(**raw)

    def test_invalid_instance_id_characters(self) -> None:
        raw = _minimal_raw()
        raw["instance_id"] = "bad_chars!"
        with pytest.raises(ValueError, match="invalid characters"):
            AppConfig(**raw)

    def test_invalid_permission_value(self) -> None:
        raw = _minimal_raw()
        raw["modules"]["filestation"]["permission"] = "superadmin"
        with pytest.raises(ValueError):
            AppConfig(**raw)

    def test_module_defaults(self) -> None:
        mod = ModuleConfig()
        assert mod.enabled is True
        assert mod.permission == "read"
        assert mod.settings == {}

    def test_connection_defaults(self) -> None:
        conn = ConnectionConfig(host="test")
        assert conn.https is False
        assert conn.verify_ssl is True
        assert conn.timeout == 30

    def test_logging_defaults(self) -> None:
        log = LoggingConfig()
        assert log.level == "info"
        assert log.file is None


class TestDeriveInstanceId:
    def test_ip_address(self) -> None:
        assert _derive_instance_id("192.168.1.100") == "192-168-1-100"

    def test_hostname(self) -> None:
        assert _derive_instance_id("nas.local") == "nas"

    def test_simple_hostname(self) -> None:
        assert _derive_instance_id("mynas") == "mynas"


class TestEnvOverrides:
    def test_host_override(self) -> None:
        raw: dict[str, Any] = {}
        with patch.dict(os.environ, {"SYNOLOGY_HOST": "10.0.0.1"}):
            result = _merge_env_overrides(raw)
        assert result["connection"]["host"] == "10.0.0.1"

    def test_port_coerced_to_int(self) -> None:
        raw: dict[str, Any] = {"connection": {}}
        with patch.dict(os.environ, {"SYNOLOGY_PORT": "5001"}):
            result = _merge_env_overrides(raw)
        assert result["connection"]["port"] == 5001

    def test_https_coerced_to_bool(self) -> None:
        raw: dict[str, Any] = {"connection": {}}
        with patch.dict(os.environ, {"SYNOLOGY_HTTPS": "true"}):
            result = _merge_env_overrides(raw)
        assert result["connection"]["https"] is True

    def test_env_does_not_override_when_not_set(self) -> None:
        raw: dict[str, Any] = {"connection": {"host": "original"}}
        env = {k: v for k, v in os.environ.items() if not k.startswith("SYNOLOGY_")}
        with patch.dict(os.environ, env, clear=True):
            result = _merge_env_overrides(raw)
        assert result["connection"]["host"] == "original"


class TestDiscoverConfigPath:
    def test_explicit_path(self, tmp_path: Path) -> None:
        config_file = tmp_path / "config.yaml"
        config_file.write_text("schema_version: 1\n")
        result = discover_config_path(str(config_file))
        assert result == config_file

    def test_explicit_path_not_found(self) -> None:
        with pytest.raises(FileNotFoundError, match="not found"):
            discover_config_path("/nonexistent/config.yaml")

    def test_env_var_path(self, tmp_path: Path) -> None:
        config_file = tmp_path / "config.yaml"
        config_file.write_text("schema_version: 1\n")
        with patch.dict(os.environ, {"SYNOLOGY_MCP_CONFIG": str(config_file)}):
            result = discover_config_path()
        assert result == config_file


class TestLoadConfig:
    def test_load_minimal(self, tmp_path: Path) -> None:
        config_file = tmp_path / "config.yaml"
        config_file.write_text(
            "schema_version: 1\n"
            "connection:\n"
            "  host: 192.168.1.100\n"
            "modules:\n"
            "  filestation:\n"
            "    enabled: true\n"
        )
        config = load_config(config_file)
        assert config.connection is not None
        assert config.connection.host == "192.168.1.100"
        assert "filestation" in config.modules



"""Tests for core/auth.py — auth flows, credential resolution, re-auth."""

from __future__ import annotations

import os
from typing import Any
from unittest.mock import MagicMock, patch

import httpx
import pytest
import respx

from synology_mcp.core.auth import AuthManager
from synology_mcp.core.client import DsmClient
from synology_mcp.core.config import AppConfig
from synology_mcp.core.errors import AuthenticationError
from synology_mcp.core.state import ApiInfoEntry

BASE_URL = "http://nas:5000"


def _make_config(**overrides: Any) -> AppConfig:
    raw: dict[str, Any] = {
        "schema_version": 1,
        "connection": {"host": "nas", "port": 5000},
        "modules": {"filestation": {"enabled": True}},
    }
    raw.update(overrides)
    return AppConfig(**raw)


def _make_client() -> DsmClient:
    client = DsmClient(base_url=BASE_URL)
    client._api_cache = {
        "SYNO.API.Auth": ApiInfoEntry(path="entry.cgi", min_version=1, max_version=7),
        "SYNO.FileStation.List": ApiInfoEntry(path="entry.cgi", min_version=1, max_version=2),
    }
    return client


def _no_keyring() -> MagicMock:
    """Return a mock keyring module where get_password raises."""
    mock = MagicMock()
    mock.get_password.side_effect = Exception("No keyring backend")
    return mock


def _keyring_with(
    username: str | None = None,
    password: str | None = None,
    device_id: str | None = None,
) -> MagicMock:
    """Return a mock keyring module that returns specific values."""
    mock = MagicMock()
    mock.get_password.side_effect = lambda _svc, key: {
        "username": username,
        "password": password,
        "device_id": device_id,
    }.get(key)
    return mock


def _clean_env() -> dict[str, str]:
    """Return env dict with all SYNOLOGY_ vars removed."""
    return {k: v for k, v in os.environ.items() if not k.startswith("SYNOLOGY_")}


class TestCredentialResolution:
    def test_credentials_from_config(self) -> None:
        config = _make_config(auth={"username": "admin", "password": "secret"})
        client = _make_client()
        auth = AuthManager(config, client)

        with (
            patch.dict(os.environ, _clean_env(), clear=True),
            patch("synology_mcp.core.auth.kr", _no_keyring()),
        ):
            username, password, device_id = auth._resolve_credentials()

        assert username == "admin"
        assert password == "secret"
        assert device_id is None

    def test_credentials_from_env(self) -> None:
        config = _make_config()
        client = _make_client()
        auth = AuthManager(config, client)

        env = {
            **_clean_env(),
            "SYNOLOGY_USERNAME": "env_user",
            "SYNOLOGY_PASSWORD": "env_pass",
            "SYNOLOGY_DEVICE_ID": "env_device",
        }
        with (
            patch.dict(os.environ, env, clear=True),
            patch("synology_mcp.core.auth.kr", _no_keyring()),
        ):
            username, password, device_id = auth._resolve_credentials()

        assert username == "env_user"
        assert password == "env_pass"
        assert device_id == "env_device"

    def test_credentials_from_keyring(self) -> None:
        config = _make_config()
        client = _make_client()
        auth = AuthManager(config, client)

        with (
            patch.dict(os.environ, _clean_env(), clear=True),
            patch(
                "synology_mcp.core.auth.kr",
                _keyring_with("kr_user", "kr_pass", "kr_device"),
            ),
        ):
            username, password, device_id = auth._resolve_credentials()

        assert username == "kr_user"
        assert password == "kr_pass"
        assert device_id == "kr_device"

    def test_no_credentials_raises(self) -> None:
        config = _make_config()
        client = _make_client()
        auth = AuthManager(config, client)

        with (
            patch.dict(os.environ, _clean_env(), clear=True),
            patch("synology_mcp.core.auth.kr", _no_keyring()),
            pytest.raises(AuthenticationError, match="No credentials"),
        ):
            auth._resolve_credentials()


class TestLogin:
    @respx.mock
    async def test_simple_login(self) -> None:
        respx.get(f"{BASE_URL}/webapi/entry.cgi").respond(
            json={"success": True, "data": {"sid": "test-sid-123"}}
        )

        config = _make_config(auth={"username": "admin", "password": "secret"})
        async with _make_client() as client:
            auth = AuthManager(config, client)
            with (
                patch.dict(os.environ, _clean_env(), clear=True),
                patch("synology_mcp.core.auth.kr", _no_keyring()),
            ):
                sid = await auth.login()

        assert sid == "test-sid-123"

    @respx.mock
    async def test_login_with_device_id(self) -> None:
        route = respx.get(f"{BASE_URL}/webapi/entry.cgi").respond(
            json={"success": True, "data": {"sid": "2fa-sid-456"}}
        )

        config = _make_config(
            auth={"username": "admin", "password": "secret", "device_id": "dev123"}
        )
        async with _make_client() as client:
            auth = AuthManager(config, client)
            with (
                patch.dict(os.environ, _clean_env(), clear=True),
                patch("synology_mcp.core.auth.kr", _no_keyring()),
            ):
                sid = await auth.login()

        assert sid == "2fa-sid-456"
        request_params = dict(route.calls[0].request.url.params)
        assert request_params["device_id"] == "dev123"

    @respx.mock
    async def test_2fa_required_without_device_id(self) -> None:
        respx.get(f"{BASE_URL}/webapi/entry.cgi").respond(
            json={"success": False, "error": {"code": 403}}
        )

        config = _make_config(auth={"username": "admin", "password": "secret"})
        async with _make_client() as client:
            auth = AuthManager(config, client)
            with (
                patch.dict(os.environ, _clean_env(), clear=True),
                patch("synology_mcp.core.auth.kr", _no_keyring()),
                pytest.raises(AuthenticationError, match="2FA"),
            ):
                await auth.login()


class TestReAuth:
    @respx.mock
    async def test_re_auth_on_session_expired(self) -> None:
        call_count = 0

        def side_effect(request: httpx.Request) -> httpx.Response:
            nonlocal call_count
            call_count += 1
            params = dict(request.url.params)
            if params.get("method") == "list_share" and call_count == 1:
                return httpx.Response(200, json={"success": False, "error": {"code": 106}})
            if params.get("method") == "login":
                return httpx.Response(200, json={"success": True, "data": {"sid": "new-sid"}})
            return httpx.Response(200, json={"success": True, "data": {"shares": []}})

        respx.get(f"{BASE_URL}/webapi/entry.cgi").mock(side_effect=side_effect)

        config = _make_config(auth={"username": "admin", "password": "secret"})
        async with _make_client() as client:
            AuthManager(config, client)
            client.sid = "old-sid"

            with (
                patch.dict(os.environ, _clean_env(), clear=True),
                patch("synology_mcp.core.auth.kr", _no_keyring()),
            ):
                data = await client.request("SYNO.FileStation.List", "list_share", version=2)
        assert "shares" in data

    @respx.mock
    async def test_get_session_returns_existing(self) -> None:
        config = _make_config(auth={"username": "admin", "password": "secret"})
        async with _make_client() as client:
            auth = AuthManager(config, client)
            client.sid = "existing-sid"
            sid = await auth.get_session()
        assert sid == "existing-sid"

    @respx.mock
    async def test_get_session_logs_in_when_no_sid(self) -> None:
        respx.get(f"{BASE_URL}/webapi/entry.cgi").respond(
            json={"success": True, "data": {"sid": "fresh-sid"}}
        )
        config = _make_config(auth={"username": "admin", "password": "secret"})
        async with _make_client() as client:
            auth = AuthManager(config, client)
            with (
                patch.dict(os.environ, _clean_env(), clear=True),
                patch("synology_mcp.core.auth.kr", _no_keyring()),
            ):
                sid = await auth.get_session()
        assert sid == "fresh-sid"


class TestCredentialPriority:
    """Test that credential resolution follows: env > config > keyring."""

    def test_env_overrides_keyring(self) -> None:
        """Env vars should take priority over keyring."""
        config = _make_config()
        client = _make_client()
        auth = AuthManager(config, client)

        env = {
            **_clean_env(),
            "SYNOLOGY_USERNAME": "env_user",
            "SYNOLOGY_PASSWORD": "env_pass",
        }
        with (
            patch.dict(os.environ, env, clear=True),
            patch(
                "synology_mcp.core.auth.kr",
                _keyring_with("kr_user", "kr_pass"),
            ),
        ):
            username, password, device_id = auth._resolve_credentials()

        assert username == "env_user"
        assert password == "env_pass"

    def test_config_overrides_keyring(self) -> None:
        """Config file creds should take priority over keyring."""
        config = _make_config(auth={"username": "cfg_user", "password": "cfg_pass"})
        client = _make_client()
        auth = AuthManager(config, client)

        with (
            patch.dict(os.environ, _clean_env(), clear=True),
            patch(
                "synology_mcp.core.auth.kr",
                _keyring_with("kr_user", "kr_pass"),
            ),
        ):
            username, password, device_id = auth._resolve_credentials()

        assert username == "cfg_user"
        assert password == "cfg_pass"

    def test_env_overrides_config(self) -> None:
        """Env vars should take priority over config file creds."""
        config = _make_config(auth={"username": "cfg_user", "password": "cfg_pass"})
        client = _make_client()
        auth = AuthManager(config, client)

        env = {
            **_clean_env(),
            "SYNOLOGY_USERNAME": "env_user",
            "SYNOLOGY_PASSWORD": "env_pass",
        }
        with (
            patch.dict(os.environ, env, clear=True),
            patch("synology_mcp.core.auth.kr", _no_keyring()),
        ):
            username, password, device_id = auth._resolve_credentials()

        assert username == "env_user"
        assert password == "env_pass"

    def test_partial_env_falls_through(self) -> None:
        """If env has username but not password, password comes from keyring."""
        config = _make_config()
        client = _make_client()
        auth = AuthManager(config, client)

        env = {
            **_clean_env(),
            "SYNOLOGY_USERNAME": "env_user",
        }
        with (
            patch.dict(os.environ, env, clear=True),
            patch(
                "synology_mcp.core.auth.kr",
                _keyring_with("kr_user", "kr_pass"),
            ),
        ):
            username, password, device_id = auth._resolve_credentials()

        assert username == "env_user"
        assert password == "kr_pass"

    def test_device_id_from_keyring_when_not_in_env(self) -> None:
        """Device ID from keyring is used even when creds come from env."""
        config = _make_config()
        client = _make_client()
        auth = AuthManager(config, client)

        env = {**_clean_env(), "SYNOLOGY_USERNAME": "env_user", "SYNOLOGY_PASSWORD": "env_pass"}
        with (
            patch.dict(os.environ, env, clear=True),
            patch(
                "synology_mcp.core.auth.kr",
                _keyring_with(device_id="kr_device"),
            ),
        ):
            username, password, device_id = auth._resolve_credentials()

        assert username == "env_user"
        assert password == "env_pass"
        assert device_id == "kr_device"


class TestDbusAutoDetect:
    def test_dbus_set_when_missing_and_socket_exists(self) -> None:
        """On Linux, auto-set DBUS_SESSION_BUS_ADDRESS if socket exists."""
        config = _make_config()
        client = _make_client()
        auth = AuthManager(config, client)

        clean = _clean_env()
        # Ensure DBUS is not set
        clean.pop("DBUS_SESSION_BUS_ADDRESS", None)

        with (
            patch.dict(os.environ, clean, clear=True),
            patch("synology_mcp.core.auth.kr", _keyring_with("user", "pass")),
            patch("sys.platform", "linux"),
            patch("pathlib.Path.exists", return_value=True),
            patch("os.getuid", return_value=1000),
        ):
            auth._resolve_credentials()

        # After resolution, DBUS should have been set (if we're on linux)
        # This test validates the code path runs without error

    def test_dbus_not_set_on_macos(self) -> None:
        """On macOS, don't set DBUS_SESSION_BUS_ADDRESS."""
        config = _make_config(auth={"username": "admin", "password": "secret"})
        client = _make_client()
        auth = AuthManager(config, client)

        clean = _clean_env()
        clean.pop("DBUS_SESSION_BUS_ADDRESS", None)

        with (
            patch.dict(os.environ, clean, clear=True),
            patch("synology_mcp.core.auth.kr", _no_keyring()),
            patch("sys.platform", "darwin"),
        ):
            username, password, _ = auth._resolve_credentials()

        assert username == "admin"


class TestSessionNaming:
    def test_session_name_format(self) -> None:
        config = _make_config(instance_id="test-nas")
        client = _make_client()
        auth = AuthManager(config, client)
        assert auth._session_name.startswith("SynologyMCP_test-nas_")
        uuid_part = auth._session_name.split("_")[-1]
        assert len(uuid_part) == 8



"""Tests for core/state.py — state file read/write."""

from __future__ import annotations

from unittest.mock import patch

from synology_mcp.core.state import ApiInfoEntry, ServerState, load_state, save_state


class TestServerState:
    def test_default_state(self) -> None:
        state = ServerState()
        assert state.api_info_cache == {}
        assert state.negotiated_versions == {}
        assert state.last_connected is None

    def test_load_nonexistent_returns_default(self, tmp_path: object) -> None:
        state = load_state("nonexistent-instance")
        assert isinstance(state, ServerState)
        assert state.api_info_cache == {}

    def test_save_and_load_roundtrip(self, tmp_path: object) -> None:
        instance_id = "test-roundtrip"
        state = ServerState(
            api_info_cache={
                "SYNO.API.Auth": ApiInfoEntry(path="entry.cgi", min_version=1, max_version=7),
            },
            negotiated_versions={"SYNO.API.Auth": 6},
            last_connected="2026-03-16T22:00:00Z",
            dsm_version="7.2.1-69057",
            hostname="MyNAS",
        )

        # Use tmp_path for home directory to avoid polluting real filesystem
        with patch("synology_mcp.core.state.Path.home", return_value=tmp_path):
            save_state(instance_id, state)
            loaded = load_state(instance_id)

        assert loaded.negotiated_versions == {"SYNO.API.Auth": 6}
        assert loaded.last_connected == "2026-03-16T22:00:00Z"
        assert loaded.dsm_version == "7.2.1-69057"
        assert "SYNO.API.Auth" in loaded.api_info_cache
        assert loaded.api_info_cache["SYNO.API.Auth"].max_version == 7



"""Tests for modules/filestation/search.py — search_files."""

from __future__ import annotations

from typing import TYPE_CHECKING

import httpx
import respx

from synology_mcp.modules.filestation.search import search_files
from tests.conftest import BASE_URL

if TYPE_CHECKING:
    from synology_mcp.core.client import DsmClient


class TestSearchFiles:
    @respx.mock
    async def test_search_success(self, mock_client: DsmClient) -> None:
        call_count = 0

        def side_effect(request: httpx.Request) -> httpx.Response:
            nonlocal call_count
            call_count += 1
            params = dict(request.url.params)
            if params.get("method") == "start":
                return httpx.Response(200, json={"success": True, "data": {"taskid": "search-1"}})
            if params.get("method") == "list":
                return httpx.Response(
                    200,
                    json={
                        "success": True,
                        "data": {
                            "files": [
                                {
                                    "name": "Severance.S02E10.mkv",
                                    "path": "/video/Downloads/Severance.S02E10.mkv",
                                    "isdir": False,
                                    "additional": {
                                        "size": 1932735283,
                                        "time": {"mtime": 1710000000},
                                    },
                                },
                            ],
                            "finished": True,
                            "total": 1,
                        },
                    },
                )
            # stop/clean
            return httpx.Response(200, json={"success": True, "data": {}})

        respx.get(f"{BASE_URL}/webapi/entry.cgi").mock(side_effect=side_effect)

        result = await search_files(mock_client, folder_path="/video", pattern="*Severance*")
        assert "Severance.S02E10.mkv" in result
        assert "1 results found" in result

    @respx.mock
    async def test_search_with_exclude_pattern(self, mock_client: DsmClient) -> None:
        call_count = 0

        def side_effect(request: httpx.Request) -> httpx.Response:
            nonlocal call_count
            call_count += 1
            params = dict(request.url.params)
            if params.get("method") == "start":
                return httpx.Response(200, json={"success": True, "data": {"taskid": "search-2"}})
            if params.get("method") == "list":
                return httpx.Response(
                    200,
                    json={
                        "success": True,
                        "data": {
                            "files": [
                                {
                                    "name": "Severance.S02E10.mkv",
                                    "path": "/video/Severance.S02E10.mkv",
                                    "isdir": False,
                                    "additional": {"size": 1932735283, "time": {"mtime": 0}},
                                },
                                {
                                    "name": "Severance.S02E10.torrent",
                                    "path": "/video/Severance.S02E10.torrent",
                                    "isdir": False,
                                    "additional": {"size": 1024, "time": {"mtime": 0}},
                                },
                            ],
                            "finished": True,
                            "total": 2,
                        },
                    },
                )
            return httpx.Response(200, json={"success": True, "data": {}})

        respx.get(f"{BASE_URL}/webapi/entry.cgi").mock(side_effect=side_effect)

        result = await search_files(
            mock_client,
            folder_path="/video",
            pattern="*Severance*",
            exclude_pattern="*.torrent",
        )
        assert "Severance.S02E10.mkv" in result
        assert "torrent" not in result.split("results")[0]  # not in the data rows
        assert "1 excluded by filter" in result

    @respx.mock
    async def test_search_zero_results(self, mock_client: DsmClient) -> None:
        call_count = 0

        def side_effect(request: httpx.Request) -> httpx.Response:
            nonlocal call_count
            call_count += 1
            params = dict(request.url.params)
            if params.get("method") == "start":
                return httpx.Response(200, json={"success": True, "data": {"taskid": "search-3"}})
            if params.get("method") == "list":
                return httpx.Response(
                    200,
                    json={
                        "success": True,
                        "data": {"files": [], "finished": True, "total": 0},
                    },
                )
            return httpx.Response(200, json={"success": True, "data": {}})

        respx.get(f"{BASE_URL}/webapi/entry.cgi").mock(side_effect=side_effect)

        result = await search_files(mock_client, folder_path="/video", pattern="nonexistent")
        assert "0 results" in result

    @respx.mock
    async def test_search_error(self, mock_client: DsmClient) -> None:
        respx.get(f"{BASE_URL}/webapi/entry.cgi").respond(
            json={"success": False, "error": {"code": 408}}
        )
        result = await search_files(mock_client, folder_path="/nonexistent")
        assert "[!]" in result

    @respx.mock
    async def test_search_with_size_filter(self, mock_client: DsmClient) -> None:
        call_count = 0

        def side_effect(request: httpx.Request) -> httpx.Response:
            nonlocal call_count
            call_count += 1
            params = dict(request.url.params)
            if params.get("method") == "start":
                # Verify size params were passed
                assert "size_from" in params
                return httpx.Response(200, json={"success": True, "data": {"taskid": "search-4"}})
            if params.get("method") == "list":
                return httpx.Response(
                    200,
                    json={
                        "success": True,
                        "data": {"files": [], "finished": True, "total": 0},
                    },
                )
            return httpx.Response(200, json={"success": True, "data": {}})

        respx.get(f"{BASE_URL}/webapi/entry.cgi").mock(side_effect=side_effect)

        await search_files(mock_client, folder_path="/video", size_from="500MB")
